#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import unittest
from datetime import datetime, timedelta

from container_app_conf import Config
from container_app_conf.entry.bool import BoolConfigEntry
from container_app_conf.entry.date import DateConfigEntry
from container_app_conf.entry.float import FloatConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.list import ListConfigEntry
from container_app_conf.entry.string import StringConfigEntry
from container_app_conf.entry.timedelta import TimeDeltaConfigEntry


class TestConfig(Config):
    @property
    def config_file_names(self) -> [str]:
        return ["testing"]

    BOOL = BoolConfigEntry(
        yaml_path=["test", "bool"],
        default=False
    )
    STRING = StringConfigEntry(
        yaml_path=["test", "string"],
        default="default value"
    )
    INT = IntConfigEntry(
        yaml_path=["test", "int"],
        default=100
    )
    FLOAT = FloatConfigEntry(
        yaml_path=["test", "float"],
        default=1.23
    )
    DATE = DateConfigEntry(
        yaml_path=["test", "this", "date", "is", "nested", "deep"],
        default=datetime.now()
    )
    TIMEDELTA = TimeDeltaConfigEntry(
        yaml_path=["test", "this", "timediff", "is", "in", "this", "branch"],
        default=timedelta(seconds=10)
    )
    STRING_LIST = ListConfigEntry(
        item_type=StringConfigEntry,
        yaml_path=["test", "this", "is", "a", "list"],
        default=[
            "these",
            "are",
            "test",
            "values"
        ]
    )


class TestBase(unittest.TestCase):
    under_test = TestConfig()



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from datetime import timedelta

from pytimeparse import parse

from container_app_conf import ConfigEntry


class TimeDeltaConfigEntry(ConfigEntry):
    _example = "4h32m1s"

    def _value_to_type(self, value: any) -> timedelta or None:
        """
        Tries to permissively convert the given value to a timedelta.
        :param value: the value to parse
        :return: the parsed date value
        """
        if isinstance(value, timedelta):
            return value
        elif isinstance(value, str):
            parsed = parse(value)
            if parsed is None:
                raise ValueError("Cannot parse the given timedelta format: {}".format(value))
            return timedelta(seconds=parsed)
        else:
            self._raise_invalid_value(value)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import logging


class ConfigEntry:
    _example = None

    def __init__(self, yaml_path: [str], example: any = None, description: str or None = None, default: any = None,
                 none_allowed: bool = None):
        """
        Creates a config entry
        :param yaml_path: list of strings representing the yaml tree path for this entry
        :param example: example str value
        :param description: a description of this entry
        :param default: the default value
        :param none_allowed: Set to True if a 'None' value may be allowed, False if not,
                             otherwise it will be True if the default value is not None.
        """
        if len(yaml_path) <= 0:
            raise ValueError("{}: yaml_path must contain at least one node".format(self.__class__.__name__))
        self.yaml_path = yaml_path
        self.env_key = "_".join(yaml_path).upper()

        self.description = description
        if example is not None:
            self._example = example

        if none_allowed is None:
            none_allowed = default is None
        self._none_allowed = none_allowed

        self.default = self._parse_value(default)
        self._value = default

    @property
    def example(self) -> str:
        return self.default if self.default is not None else self._example

    @property
    def value(self) -> any:
        """
        :return: the value of this config entry
        """
        return self._value

    @value.setter
    def value(self, new_value) -> None:
        """
        :param new_value: the new value to set
        """
        self._value = self._parse_value(new_value)

    def _parse_value(self, value: any) -> any or None:
        """
        Tries to permissively convert the given value to the expected value type.
        :param value: the value to parse
        :return: the parsed value
        """
        if value is None:
            if self._none_allowed:
                return None
            else:
                self._raise_invalid_value(value)

        try:
            return self._value_to_type(value)
        except Exception as ex:
            logging.exception(ex)
            self._raise_invalid_value(value)

    def _value_to_type(self, value: any) -> any:
        """
        Converts the given value to the expected value type of this entry
        :param value: the yaml value
        :return: parsed value
        """
        raise NotImplementedError()

    def _type_to_value(self, type: any) -> any:
        """
        Converts a value of the expected entry type to a valid representation in the config file.
        This is the inverse function of _value_to_type
        :param type: value of expected entry type
        :return: config file value
        """
        return str(type)

    def _raise_invalid_value(self, value: any):
        raise ValueError("Invalid value '{}' for config option {}".format(value, self.env_key))



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from container_app_conf import ConfigEntry


class IntConfigEntry(ConfigEntry):
    _example = "42"

    def _value_to_type(self, value: any) -> int or None:
        """
        Tries to permissively convert the given value to an int.
        :param value: the value to parse
        :return: the parsed int value
        """
        return int(value)

    def _type_to_value(self, type: any) -> any:
        return int(type)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from container_app_conf import ConfigEntry


class StringConfigEntry(ConfigEntry):
    _example = "text"

    def _value_to_type(self, value: any) -> str or None:
        """
        Converts the given type to the expected type
        :param value: the yaml value
        :return: parsed value
        """
        s = str(value)
        if self._none_allowed:
            if s.lower() in ['none', 'null', 'nil']:
                return None

        return s



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from container_app_conf.entry.bool import BoolConfigEntry
from container_app_conf.entry.date import DateConfigEntry
from container_app_conf.entry.float import FloatConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.list import ListConfigEntry
from container_app_conf.entry.timedelta import TimeDeltaConfigEntry
from tests import TestBase


class EntryTest(TestBase):

    @staticmethod
    def test_bool_entry():
        bool_config_entry = BoolConfigEntry(yaml_path=["bool"])

        true_values = ["y", "yes", "true", "t", 1, True]
        false_values = ["n", "no", "false", "f", 0, False]
        invalid_values = ["hello", 2, 0.1]

        for v in true_values:
            assert bool_config_entry._parse_value(v)

        for v in false_values:
            assert not bool_config_entry._parse_value(v)

        for v in invalid_values:
            try:
                bool_config_entry._parse_value(v)
                assert False
            except ValueError as ex:
                assert ex is not None

    @staticmethod
    def test_int_entry():
        int_config_entry = IntConfigEntry(yaml_path=["int"])

        input_result_map = {
            "5": 5,
            5: 5,
            "-3": -3,
            -3: -3
        }

        for k, v in input_result_map.items():
            assert int_config_entry._parse_value(k) == v

    @staticmethod
    def test_float_entry():
        float_config_entry = FloatConfigEntry(yaml_path=["float"])

        input_result_map = {
            "5": 5.0,
            5: 5.0,
            "-3.0": -3.0,
            -3.0: -3.0,
            1.2: 1.2,
            "1.2": 1.2,
            "3%": 0.03
        }

        for k, v in input_result_map.items():
            assert float_config_entry._parse_value(k) == v

    @staticmethod
    def test_int_list_entry():
        int_list_config_entry = ListConfigEntry(item_type=IntConfigEntry, yaml_path=["int_list"])

        input_result_map = {
            "5": [5],
            "": [],
            "1,2,3": [1, 2, 3]
        }

        for k, v in input_result_map.items():
            assert int_list_config_entry._parse_value(k) == v

    @staticmethod
    def test_float_list_entry():
        int_list_config_entry = ListConfigEntry(item_type=FloatConfigEntry, yaml_path=["float_list"])

        input_result_map = {
            "5": [5.0],
            "": [],
            "1,2.5,3": [1.0, 2.5, 3.0]
        }

        for k, v in input_result_map.items():
            assert int_list_config_entry._parse_value(k) == v

    @staticmethod
    def test_date_entry():
        from datetime import datetime
        from dateutil.tz import tzutc

        date_config_entry = DateConfigEntry(yaml_path=["date"])
        input_result_map = {
            "2008-09-03T20:56:35.450686Z": datetime(2008, 9, 3, 20, 56, 35, 450686, tzinfo=tzutc()),
            "2008-09-03": datetime(2008, 9, 3, 0, 0, 0, 0),
        }

        for k, v in input_result_map.items():
            assert date_config_entry._parse_value(k) == v

    @staticmethod
    def test_timedelta_entry():
        from datetime import timedelta

        date_config_entry = TimeDeltaConfigEntry(yaml_path=["timedelta"])

        input_result_map = {
            "20:56:35": timedelta(hours=20, minutes=56, seconds=35),
            "32m": timedelta(minutes=32),
            "4h0m3s": timedelta(hours=4, minutes=0, seconds=3),
            "4h3s": timedelta(hours=4, minutes=0, seconds=3),
            "4:13": timedelta(hours=0, minutes=4, seconds=13),
        }

        for k, v in input_result_map.items():
            assert date_config_entry._parse_value(k) == v



#  Copyright (c) 2019 Markus Ressel
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
import os
import subprocess

from setuptools import setup, find_packages

VERSION_NUMBER = "1.2.1"

GIT_BRANCH = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
GIT_BRANCH = GIT_BRANCH.decode()  # convert to standard string
GIT_BRANCH = GIT_BRANCH.rstrip()  # remove unnecessary whitespace

if GIT_BRANCH == "master":
    DEVELOPMENT_STATUS = "Development Status :: 5 - Production/Stable"
    VERSION_NAME = VERSION_NUMBER
elif GIT_BRANCH == "beta":
    DEVELOPMENT_STATUS = "Development Status :: 4 - Beta"
    VERSION_NAME = "%s-beta" % VERSION_NUMBER
elif GIT_BRANCH == "dev":
    DEVELOPMENT_STATUS = "Development Status :: 3 - Alpha"
    VERSION_NAME = "%s-dev" % VERSION_NUMBER
elif os.environ.get("TRAVIS_BRANCH", None) == os.environ.get("TRAVIS_TAG", None) == "v{}".format(VERSION_NUMBER):
    # travis tagged release branch
    DEVELOPMENT_STATUS = "Development Status :: 5 - Production/Stable"
    VERSION_NAME = VERSION_NUMBER
else:
    print("Unknown git branch, using pre-alpha as default")
    DEVELOPMENT_STATUS = "Development Status :: 2 - Pre-Alpha"
    VERSION_NAME = "%s-%s" % (VERSION_NUMBER, GIT_BRANCH)


def readme_type() -> str:
    import os
    if os.path.exists("README.rst"):
        return "text/x-rst"
    if os.path.exists("README.md"):
        return "text/markdown"


def readme() -> [str]:
    if readme_type() == "text/markdown":
        file_name = "README.md"
    else:
        file_name = "README.rst"

    with open(file_name) as f:
        return f.read()


def install_requirements() -> [str]:
    return read_requirements_file("requirements.txt")


def test_requirements() -> [str]:
    return read_requirements_file("test_requirements.txt")


def read_requirements_file(file_name: str):
    with open(file_name, encoding='utf-8') as f:
        requirements_file = f.readlines()
    return [r.strip() for r in requirements_file]


setup(
    name='container_app_conf',
    version=VERSION_NAME,
    description='Convenient configuration of containerized applications',
    long_description=readme(),
    long_description_content_type=readme_type(),
    license='GPLv3+',
    author='Markus Ressel',
    author_email='mail@markusressel.de',
    url='https://github.com/markusressel/container-app-conf',
    packages=find_packages(),
    classifiers=[
        DEVELOPMENT_STATUS,
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    install_requires=install_requirements(),
    tests_require=test_requirements()
)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from typing import Type

from container_app_conf import ConfigEntry


class ListConfigEntry(ConfigEntry):
    """
    Config entry allowing to specify a list of items of specified type
    """

    def __init__(self, item_type: Type[ConfigEntry], yaml_path: [str], example: str = None,
                 description: str or None = None,
                 default: any = None, none_allowed: bool = None):
        """

        :param item_type: the type of individual list items
        :param yaml_path: list of yaml tree entries
        :param example: example str value
        :param description: a description of this list entry
        :param default: default value
        :param none_allowed: if None is allowed for this config entry
        """
        self._item_type = item_type

        super().__init__(
            yaml_path=yaml_path,
            description=description,
            example=example,
            default=default,
            none_allowed=none_allowed
        )

    @property
    def example(self) -> str:
        if self.default is not None:
            if isinstance(self.default, list):
                return ",".join(self.default)
            else:
                return self.default

        single_example = self._item_type("dummy").example
        return ",".join([single_example, single_example, single_example])

    def _value_to_type(self, value: any) -> [any] or None:
        """
        Tries to permissively convert the given value to a list of values.
        :param value: the value to parse
        :return: the parsed list
        """
        if not isinstance(value, list):
            value = str(value).split(',')

        return list(map(lambda x: self._item_type._value_to_type(self, x), filter(lambda x: x, value)))

    def _type_to_value(self, type: list or str) -> any:
        if isinstance(type, str):
            return type
        str_items = list(map(lambda x: self._item_type._type_to_value(self, x), type))
        return ",".join(str_items)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import logging
import os

import yaml

from container_app_conf.const import DEFAULT_CONFIG_FILE_PATHS
from container_app_conf.entry import ConfigEntry

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


class Config:
    """
    Config base class.
    Extend this in your application.
    """

    def __init__(self, validate: bool = True):
        """
        Creates a config object and reads configuration.
        :param validate: if validation should be run (can be disabled for tests)
        """
        self._config_entries = self._find_config_entries()

        if self._find_config_file() is None:
            self.write_reference_yaml()

        self._read_yaml()
        self._read_env()
        if validate:
            self.validate()

    @property
    def config_file_names(self) -> [str]:
        """
        :return: List of allowed config file names
        """
        raise NotImplementedError()

    @property
    def config_file_extensions(self) -> [str]:
        """
        :return: List of allowed config file extensions
        """
        return ['yaml', 'yml']

    @property
    def config_file_paths(self) -> [str]:
        """
        :return: List of allowed config file paths
        """
        return DEFAULT_CONFIG_FILE_PATHS

    def validate(self):
        """
        Validates the current configuration and throws an exception if something is wrong
        """
        return

    def write_reference_yaml(self):
        """
        Writes a reference config file
        :return:
        """
        reference = self.generate_reference_config()
        text = yaml.dump(reference)

        folder = self.config_file_paths[0]
        file_name = "{}_reference.{}".format(self.config_file_names[0], self.config_file_extensions[0])
        file_path = os.path.join(self.config_file_paths[0], file_name)

        os.makedirs(folder, exist_ok=True)
        with open(file_path, "w") as file:
            file.write(text)

    def generate_reference_config(self) -> {}:
        """
        Generates a dictionary containing the expected config tree filled with default and example values
        :return: a dictionary containing the expected config tree
        """
        entries = self._find_config_entries()

        config_tree = {}
        for entry in entries:
            current_level = config_tree
            for path in entry.yaml_path[:-1]:
                if path not in current_level:
                    current_level[path] = {}
                current_level = current_level[path]

            current_level[entry.yaml_path[-1]] = entry._type_to_value(entry.example)

        return config_tree

    def _find_config_entries(self) -> [ConfigEntry]:
        """
        Detects config entry constants in this class
        :return: list of config entries
        """
        entries = set()
        for attribute in map(lambda x: getattr(self, x), dir(self)):
            if isinstance(attribute, ConfigEntry):
                entries.add(attribute)

        return list(entries)

    def _find_config_file(self) -> str or None:
        """
        Tries to find a usable config file
        :return: file path or None
        """
        import os

        for path in self.config_file_paths:
            path = os.path.expanduser(path)
            for extension in self.config_file_extensions:
                for file_name in self.config_file_names:
                    file_path = os.path.join(path, "{}.{}".format(file_name, extension))
                    if os.path.isfile(file_path):
                        return file_path

        return None

    def _read_yaml(self) -> None:
        """
        Reads configuration parameters from a yaml config file (if it exists)
        """

        def _get_value(root: {}, config_entry: ConfigEntry) -> any or None:
            """
            Helper function to read the value of a given config entry
            :param root: the yaml root node dictionary
            :param config_entry: the config entry to search for
            :return: the found value or the config entry's current value
            """
            value = root
            for key in config_entry.yaml_path:
                value = value.get(key)
                if value is None:
                    return config_entry.value

            return value

        file_path = self._find_config_file()
        if file_path is None:
            LOGGER.debug("No config file found for path: {}".format(file_path))
            return

        with open(file_path, 'r') as ymlfile:
            import yaml
            cfg = yaml.load(ymlfile, Loader=yaml.FullLoader)
            if not cfg:
                return

            for config_entry in self._config_entries:
                config_entry.value = _get_value(cfg, config_entry)

    def _read_env(self) -> None:
        """
        Reads configuration parameters from environment variables
        """
        import os

        for entry in self._config_entries:
            entry.value = os.environ.get(entry.env_key, entry.value)
